# ---
# title: Quick test
# output:
#   ioslides_presentation:
#     widescreen: true
#     smaller: true
# editor_options:
#   chunk_output_type: console
# jupyter:
#   jupytext:
#     cell_markers: '{{{,}}}'
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

1+2+3


