-- ---
-- jupyter:
--   kernelspec:
--     display_name: Haskell
--     language: haskell
--     name: haskell
-- ---

-- # Example Haskell Notebook

-- Define a function to add two numbers.

f :: Num a => a -> a -> a
f x y = x + y

-- Try to use the function

f 1 2 
